/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg__do_cmd_resolve2.c
 *
 * @details Implements the resolve command.
 *
 */


/*
**
** Includes
**
*/

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"


/*
**
** Globals
**
*/

/**
 * String used as a single indent by the __indent helpers.
 */
static const char* gszIndent = "  ";

/**
 * String used as the path for an item that no longer exists.
 */
static const char* gszNonExistentPath = "(non-existent)";


/*
**
** Helpers
**
*/

/**
 * Prints an indented formatted line to STDOUT.
 */
static void _resolve__stdout(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndent,  //< [in] Number of times to indent the string.
	const char* szFormat, //< [in] Format string to print, or NULL to only print the indent.
	...                   //< [in] Data to substitute into format string.
	)
{
	while (uIndent > 0u)
	{
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "%s", gszIndent)  );
		--uIndent;
	}

	if (szFormat != NULL)
	{
		va_list pArgs;

		va_start(pArgs, szFormat);
		SG_console__va(pCtx, SG_CS_STDOUT, szFormat, pArgs);
		va_end(pArgs);
		SG_ERR_CHECK_CURRENT;
	}

	SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "\n")  );

fail:
	return;
}

/**
 * Prompts the user for a line of input and returns it.
 */
static void _resolve__stdin(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndent,  //< [in] Number of times to indent the prompt.
	SG_string** ppInput,  //< [out] The input entered by the user.
	const char* szPrompt, //< [in] A prompt to display to the user, or NULL to not bother.
	...                   //< [in] Data to substitute into the prompt.
	)
{
	SG_string* sInput = NULL;

	SG_NULLARGCHECK(ppInput);

	while (uIndent > 0u)
	{
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "%s", gszIndent)  );
		--uIndent;
	}

	if (szPrompt != NULL)
	{
		va_list pArgs;

		va_start(pArgs, szPrompt);
		SG_console__va(pCtx, SG_CS_STDOUT, szPrompt, pArgs);
		va_end(pArgs);
		SG_ERR_CHECK_CURRENT;
	}

	// Note: Unlike almost every other SG function that takes an object
	//       double pointer, this one expects that we've already allocated
	//       it, rather than allocating it itself and returning it to us.
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sInput)  );
	SG_ERR_CHECK(  SG_console__readline_stdin(pCtx, &sInput)  );

	*ppInput = sInput;
	sInput = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sInput);
	return;
}

/**
 * A thread-safe and SG-like implementation of strtok.
 * TODO: Move this to sg_str_utils and write tests for it.
 */
static void _SG_strtok(
	SG_context*  pCtx,         //< [in] [out] Error and context info.
	char*        szString,     //< [in] The string to tokenize.
	                           //<      If NULL, tokenization is continued on ppHandle.
	const char*  szDelimiters, //< [in] A string of characters that will delimit the next token.
	                           //<      If NULL, then whitespace delimiters will be used.
	SG_bool      bSkipEmpty,   //< [in] If true, zero-length tokens will be skipped instead of returned.
	const char** ppToken,      //< [out] The next token in the string.
	                           //<       NULL if there are no more tokens.
	void**       ppHandle      //< [in] [out] A handle to this tokenization string/session/instance.
	)
{
	static const char* szWhitespaceDelimiters = " \t\r\n";

	const char* szToken = NULL;

	SG_NULLARGCHECK(ppToken);
	SG_NULLARGCHECK(ppHandle);

	// if they didn't supply a string, start with the handle
	if (szString == NULL)
	{
		szString = (char*)*ppHandle;
	}

	// if they didn't supply delimiters, use whitespace
	if (szDelimiters == NULL)
	{
		szDelimiters = szWhitespaceDelimiters;
	}

	// if we have a string to search, do it
	// we won't if the handle is NULL, meaning we found the last token last time
	if (szString != NULL)
	{
		// if we're skipping empty tokens
		// then advance until we find a non-delimiter
		while (
			   *szString != '\0'
			&& bSkipEmpty == SG_TRUE
			&& strchr(szDelimiters, *szString) != NULL
			)
		{
			++szString;
		}

		// store off the start of the string
		// so we can return it as the token
		szToken = szString;

		// advance until we find a delimiter
		while (
			   *szString != '\0'
			&& strchr(szDelimiters, *szString) == NULL
			)
		{
			++szString;
		}

		// setup the handle to keep going with the next call
		if (*szString == '\0')
		{
			// we found the end of the string, so this is the last token
			// set the handle to NULL, so next time we'll know that we're done
			*ppHandle = NULL;
		}
		else
		{
			// set the handle to the character after the delimiter
			// that's where we'll pick up next time
			*ppHandle = (void*)(szString + 1u);
		}

		// replace the delimiter with a NULL-terminator
		*szString = '\0';
	}

	// return the token
	*ppToken = szToken;

fail:
	return;
}


/*
**
** Private Functionality - Data Parsers
**
*/

/**
 * Parses a path specified by the user into the GID of an item.
 * The input path can be an absolute path, relative path, repo path, or NULL.
 */
static void _resolve__parse__path(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree* pPendingTree, //< [in] Pendingtree to use to convert repo paths into GIDs.
	const char*     szPath,       //< [in] The path to parse, or NULL if the user didn't specify a path.
	char**          ppGid         //< [out] The parsed repo path, or NULL if szPath is NULL.
	)
{
	SG_pathname* pAbsolutePath = NULL;
	char*        szGid         = NULL;

	SG_NULLARGCHECK(pPendingTree);
	SG_NULLARGCHECK(szPath);
	SG_NULLARGCHECK(ppGid);

	if (szPath != NULL)
	{
		if (szPath[0] != '\0' && szPath[0] == '@' && szPath[1] == '/')
		{
			// if it's a repo path, we need to first convert it to an absolute path
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, szPath, &pAbsolutePath)  );
		}
		else
		{
			// otherwise just allocate whatever they gave us as a path
			// pathnames are always absolute
			// it will interpret the path as relative to the CWD if necessary
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pAbsolutePath, szPath)  );
		}

		// get the GID of the item at the path
		SG_ERR_CHECK(  SG_pendingtree__get_gid_from_local_path(pCtx, pPendingTree, pAbsolutePath, &szGid)  );
	}

	// return the GID
	*ppGid = szGid;
	szGid = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pAbsolutePath);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Parses a string into a boolean value.
 *
 * The following strings are considered true, anything else is considered false:
 * yes, y, true, t, on, 1
 */
static void _resolve__parse__bool(
	SG_context* pCtx,   //< [in] [out] Error and context info.
	const char* szBool, //< [in] The string to parse.
	SG_bool*    pBool   //< [out] The parsed boolean value.
	)
{
	SG_NULLARGCHECK(szBool);
	SG_NULLARGCHECK(pBool);

	if (
		   SG_stricmp(szBool, "yes")  == 0
		|| SG_stricmp(szBool, "y")    == 0
		|| SG_stricmp(szBool, "true") == 0
		|| SG_stricmp(szBool, "t")    == 0
		|| SG_stricmp(szBool, "on")   == 0
		|| SG_stricmp(szBool, "1")    == 0
		)
	{
		*pBool = SG_TRUE;
	}
	else
	{
		*pBool = SG_FALSE;
	}

fail:
	return;
}

/**
 * Gets the resolve item and choice that corresond to a given path and state.
 * Prints error messages and returns NULLs if the item/choice cannot be found.
 */
static void _resolve__find_choice(
	SG_context*          pCtx,      //< [in] [out] Error and context info.
	SG_uint32            uIndent,   //< [in] Base amount to indent error messages.
	SG_resolve*          pResolve,  //< [in] Resolve data to find the item/choice in.
	const char*          szGid,     //< [in] GID of the item to find.
	SG_resolve__state    eState,    //< [in] State of the choice to find on the item.
	                                //<      If SG_RESOLVE__STATE__COUNT and the item only has one choice,
	                                //<      then that choice is returned, otherwise an error will be thrown.
	SG_resolve__item**   ppItem,    //< [out] The found item, or NULL if not found.
	SG_resolve__choice** ppChoice   //< [out] The found choice, or NULL if not found.
	)
{
	SG_resolve__item*   pItem   = NULL;
	SG_resolve__choice* pChoice = NULL;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(szGid);

	SG_ERR_CHECK(  SG_resolve__get_item__gid(pCtx, pResolve, szGid, &pItem)  );
	if (eState == SG_RESOLVE__STATE__COUNT)
	{
		// they want us to find the item's only choice
		SG_int32 iState = 0u;

		// run through each possible state
		for (iState = 0u; iState < SG_RESOLVE__STATE__COUNT; ++iState)
		{
			SG_resolve__choice* pCurrentChoice = NULL;

			// see if the item has a choice for this state
			SG_ERR_CHECK(  SG_resolve__item__check_choice__state(pCtx, pItem, (SG_resolve__state)iState, &pCurrentChoice)  );
			if (pCurrentChoice != NULL)
			{
				if (pChoice == NULL)
				{
					// this is the first choice we've found on the item
					// store it for later and possible returning
					pChoice = pCurrentChoice;
				}
				else
				{
					// this is the second choice we've found on the item
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "The specified item has multiple conflicts, so one must be explicitly specified.")  );
				}
			}
		}

		// if we didn't find a choice, throw an error
		if (pChoice == NULL)
		{
			SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "The specified item has no conflicts.")  );
		}
	}
	else
	{
		// they want a specific choice on the item
		SG_ERR_CHECK(  SG_resolve__item__get_choice__state(pCtx, pItem, eState, &pChoice)  );

		// if we didn't find a choice, throw an error
		if (pChoice == NULL)
		{
			SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "The specified conflict wasn't found on the given item.")  );
		}
	}

	// return our data
	if (ppItem != NULL)
	{
		*ppItem   = pItem;
	}
	if (ppChoice != NULL)
	{
		*ppChoice = pChoice;
	}

fail:
	return;
}


/*
**
** Private Functionality - Data Writers
**
*/

/**
 * An SG_resolve__item__callback that writes basic information about the item to stdout.
 */
static void _resolve__write_choice__item(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The current item.
	void*             pUserData, //< [in] Pointer to a base indent level (SG_uint32).
	                             //<      NULL is equivalent to zero.
	SG_bool*          pContinue  //< [out] Whether or not iteration should continue.
	)
{
	SG_uint32*             pData              = (SG_uint32*)pUserData;
	SG_uint32              uIndent            = 0u;
	SG_treenode_entry_type eType              = SG_TREENODEENTRY_TYPE__INVALID;
	const char*            szType             = NULL;
	const char*            szWorkingRepoPath  = NULL;
	const char*            szBaselineRepoPath = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pContinue);

	// if they passed an indent, use it
	if (pData != NULL)
	{
		uIndent = *pData;
	}

	// get data from the item
	SG_ERR_CHECK(  SG_resolve__item__get_type(pCtx, pItem, &eType)  );
	szType = SG_treenode_entry_type__type_to_label(eType);
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szWorkingRepoPath)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__baseline(pCtx, pItem, &szBaselineRepoPath)  );
	if (szWorkingRepoPath == NULL)
	{
		szWorkingRepoPath = gszNonExistentPath;
	}
	if (szBaselineRepoPath == NULL)
	{
		szBaselineRepoPath = gszNonExistentPath;
	}

	// write data about the item
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "%s: %s", szType, szWorkingRepoPath)  );
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Baseline Path: %s", szBaselineRepoPath)  );

fail:
	return;
}

/**
 * An SG_resolve__value__callback that retrieves information about the value.
 */
static void _resolve__write_choice__get_value(
	SG_context*        pCtx,      //< [in] [out] Error and context info.
	SG_resolve__value* pValue,    //< [in] The current value.
	void*              pUserData, //< [in] SG_vhash to add the value label/summary to.
	SG_bool*           pContinue  //< [out] Whether or not iteration should continue.
	)
{
	SG_vhash*   pData    = (SG_vhash*)pUserData;
	const char* szLabel  = NULL;
	SG_string*  sSummary = NULL;
	SG_bool     bLeaf    = SG_FALSE;
	SG_string*  sLabel   = NULL;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// get data from the value
	SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pValue, &szLabel)  );
	SG_ERR_CHECK(  SG_resolve__value__get_summary(pCtx, pValue, &sSummary)  );
	SG_ERR_CHECK(  SG_resolve__value__get_leaf(pCtx, pValue, &bLeaf)  );

	// if the value is a leaf, generate an annotated label
	if (bLeaf == SG_TRUE)
	{
		SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sLabel, "%s*", szLabel)  );
		szLabel = SG_string__sz(sLabel);
	}

	// add the data to the hash
	if (sSummary == NULL)
	{
		SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pData, szLabel)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pData, szLabel, SG_string__sz(sSummary))  );
	}

fail:
	SG_STRING_NULLFREE(pCtx, sSummary);
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

/**
 * An SG_resolve__choice__callback that writes information about the choice to stdout.
 */
static void _resolve__write_choice(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The current choice.
	SG_resolve__item*   pItem,   //< [in] The item that owns the choice, if known.
	                             //<      If NULL, this function will look it up.
	SG_uint32           uIndent  //< [in] Base level to indent the display.
	)
{
	static const char* aValueHeaders[] =
	{
		"Possible Existence States:",
		"Possible Names:",
		"Possible Locations: (parent folders)",
		"Possible Attributes:",
		"Possible Extended Attributes:",
		"Possible Contents: (use 'diff' to examine, * indicates merge leaf)",
	};

	SG_treenode_entry_type           eItemType          = SG_TREENODEENTRY_TYPE__INVALID;
	const char*                      szItemType         = NULL;
	const char*                      szWorkingRepoPath  = NULL;
	const char*                      szBaselineRepoPath = NULL;
	SG_resolve__state                eState             = SG_RESOLVE__STATE__COUNT;
	const char*                      szStateName        = NULL;
	const char*                      szStateDescription = NULL;
	SG_mrg_cset_entry_conflict_flags eConflictCauses    = SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__ZERO;
	SG_bool                          bCollisionCause    = SG_FALSE;
	SG_vhash*                        pCauseDescriptions = NULL;
	SG_uint32                        uCount             = 0u;
	SG_uint32                        uIndex             = 0u;
	SG_vhash*                        pValues            = NULL;
	SG_uint32                        uLabelWidth        = 0u;
	SG_string*                       sLabel             = NULL;
	SG_resolve__value*               pAcceptedValue     = NULL;

	SG_NULLARGCHECK(pChoice);

	// if we weren't given an item, look it up
	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__choice__get_item(pCtx, pChoice, &pItem)  );
	}

	// get data about the item
	SG_ERR_CHECK(  SG_resolve__item__get_type(pCtx, pItem, &eItemType)  );
	szItemType = SG_treenode_entry_type__type_to_label(eItemType);
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szWorkingRepoPath)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__baseline(pCtx, pItem, &szBaselineRepoPath)  );
	if (szWorkingRepoPath == NULL)
	{
		szWorkingRepoPath = gszNonExistentPath;
	}
	if (szBaselineRepoPath == NULL)
	{
		szBaselineRepoPath = gszNonExistentPath;
	}

	// get data about the choice
	SG_ERR_CHECK(  SG_resolve__choice__get_state(pCtx, pChoice, &eState)  );
	SG_ERR_CHECK(  SG_resolve__state__get_info(pCtx, eState, &szStateName, &szStateDescription)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_causes(pCtx, pChoice, &eConflictCauses, &bCollisionCause, &pCauseDescriptions)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );

	// header
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "%s %s conflict on %s: %s", (pAcceptedValue == NULL ? "Unresolved" : "Resolved"), szStateName, szItemType, szWorkingRepoPath)  );
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Baseline Path: %s", szBaselineRepoPath)  );
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Problem: %s", szStateDescription)  );

	// causes
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Cause(s):")  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pCauseDescriptions, &uCount)  );
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char* szCauseName = NULL;
		const char* szCauseDescription = NULL;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair__sz(pCtx, pCauseDescriptions, uIndex, &szCauseName, &szCauseDescription)  );
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 2u, "%s: %s", szCauseName, szCauseDescription)  );
	}

	// cycle
	if ((eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE) == SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE)
	{
		const char* szCyclePath  = NULL;
		SG_uint32   uItemsIndent = uIndent + 2u;

		SG_ERR_CHECK(  SG_resolve__choice__get_cycle_info(pCtx, pChoice, &szCyclePath)  );
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Cycling Path: %s", szCyclePath)  );
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Other Cycling Items:")  );
		SG_ERR_CHECK(  SG_resolve__choice__foreach_cycling_item(pCtx, pChoice, _resolve__write_choice__item, (void*)&uItemsIndent)  );
	}

	// collision
	if (bCollisionCause == SG_TRUE)
	{
		SG_uint32   uItemsIndent = uIndent + 2u;

		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Other Colliding Items:")  );
		SG_ERR_CHECK(  SG_resolve__choice__foreach_colliding_item(pCtx, pChoice, _resolve__write_choice__item, (void*)&uItemsIndent)  );
	}

	// values
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "%s", aValueHeaders[eState])  );
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pValues)  );
	SG_ERR_CHECK(  SG_resolve__choice__foreach_value(pCtx, pChoice, SG_FALSE, _resolve__write_choice__get_value, (void*)pValues)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pValues, &uCount)  );
	uLabelWidth = 0u;
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char* szLabel = NULL;
		SG_uint32   uLabel  = 0u;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pValues, uIndex, &szLabel, NULL)  );
		uLabel = SG_STRLEN(szLabel);
		uLabelWidth = SG_MAX(uLabelWidth, uLabel);
	}
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char*       szLabel  = NULL;
		const SG_variant* pSummary = NULL;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pValues, uIndex, &szLabel, &pSummary)  );
		if (pSummary->type == SG_VARIANT_TYPE_NULL)
		{
			SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 2u, "%s", szLabel)  );
		}
		else
		{
			SG_ASSERT(pSummary->type == SG_VARIANT_TYPE_SZ);
			SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sLabel, "%s:", szLabel)  );
			SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 2u, "%-*s %s", uLabelWidth + 1u, SG_string__sz(sLabel), pSummary->v.val_sz)  );
			SG_STRING_NULLFREE(pCtx, sLabel);
		}
	}

	// resolution
	if (pAcceptedValue != NULL)
	{
		const char* szAcceptedLabel = NULL;

		SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pAcceptedValue, &szAcceptedLabel)  );
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Resolution: %s", szAcceptedLabel)  );
	}

	// empty line
	SG_ERR_CHECK(  _resolve__stdout(pCtx, 0u, NULL)  );

fail:
	SG_VHASH_NULLFREE(pCtx, pCauseDescriptions);
	SG_VHASH_NULLFREE(pCtx, pValues);
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

/**
 * Writes the result of an accept to stdout.
 */
static void _resolve__write_accept(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_uint32           uIndent, //< [in] Base amount to indent the messages.
	SG_resolve__value*  pValue,  //< [in] The value that was accepted.
	SG_resolve__choice* pChoice, //< [in] The choice that owns the value.
	                             //<      If NULL, this function will look it up from the value.
	SG_resolve__item*   pItem    //< [in] The item that owns the choice.
	                             //<      If NULL, this function will look it up from the choice.
	)
{
	SG_treenode_entry_type eType      = SG_TREENODEENTRY_TYPE__INVALID;
	const char*            szType     = NULL;
	const char*            szRepoPath = NULL;
	SG_resolve__state      eState     = SG_RESOLVE__STATE__COUNT;
	const char*            szState    = NULL;
	const char*            szLabel    = NULL;

	SG_NULLARGCHECK(pValue);

	if (pChoice == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__value__get_choice(pCtx, pValue, &pChoice)  );
	}
	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__choice__get_item(pCtx, pChoice, &pItem)  );
	}

	SG_ERR_CHECK(  SG_resolve__item__get_type(pCtx, pItem, &eType)  );
	szType = SG_treenode_entry_type__type_to_label(eType);
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szRepoPath)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_state(pCtx, pChoice, &eState)  );
	SG_ERR_CHECK(  SG_resolve__state__get_info(pCtx, eState, &szState, NULL)  );
	SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pValue, &szLabel)  );

	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Accepted '%s' value for '%s' conflict on %s:", szLabel, szState, szType)  );
	SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "%s", szRepoPath)  );

fail:
	return;
}

/**
 * Writes the result of a merge to stdout.
 *
 * TODO: Add a message about overwriting an old value?
 */
static void _resolve__write_merge(
	SG_context*         pCtx,           //< [in] [out] Error and context info.
	SG_uint32           uIndent,        //< [in] Base amount to indent the messages.
	SG_resolve__value*  pBaselineValue, //< [in] The baseline value that was merged.
	SG_resolve__value*  pOtherValue,    //< [in] The other value that was merged.
	SG_int32            iResult,        //< [in] The result status from the merge.
	SG_resolve__value*  pResultValue,   //< [in] The result value from the merge.
	SG_resolve__choice* pChoice,        //< [in] The choice that owns the values.
	                                    //<      If NULL, this function will look it up from the values.
	SG_resolve__item*   pItem           //< [in] The item that owns the choice.
	                                    //<      If NULL, this function will look it up from the choice.
	)
{
	const char*        szResultLabel  = NULL;
	const char*        szResultTool   = NULL;
	SG_resolve__value* pAcceptedValue = NULL;

	SG_NULLARGCHECK(pBaselineValue);
	SG_NULLARGCHECK(pOtherValue);

	// if they didn't supply a choice, look it up
	if (pChoice == NULL)
	{
		SG_resolve__choice* pBaselineChoice = NULL;
		SG_resolve__choice* pOtherChoice    = NULL;

		SG_ERR_CHECK(  SG_resolve__value__get_choice(pCtx, pBaselineValue, &pBaselineChoice)  );
		SG_ERR_CHECK(  SG_resolve__value__get_choice(pCtx, pOtherValue, &pOtherChoice)  );
		SG_ARGCHECK(pBaselineChoice == pOtherChoice, pBaselineValue|pOtherValue);
		pChoice = pBaselineChoice;
	}

	// if they didn't supply an item, look it up
	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__choice__get_item(pCtx, pChoice, &pItem)  );
	}

	// check what happened and display appropriate output
	if (pResultValue != NULL)
	{
		SG_bool bMerged = SG_FALSE;

		SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pResultValue, &szResultLabel)  );
		SG_ERR_CHECK(  SG_resolve__value__get_merge(pCtx, pResultValue, &bMerged, NULL, NULL, NULL, &szResultTool, NULL)  );
		SG_ASSERT(bMerged == SG_TRUE);
	}
	switch (iResult)
	{
	case SG_MERGETOOL__STATUS__SUCCESSFUL:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Used merge tool '%s' to create value '%s' with no conflicts.", szResultTool, szResultLabel)  );
		break;
	case SG_MERGETOOL__STATUS__CONFLICT:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Used merge tool '%s' to create merged value '%s' with some conflicts remaining.", szResultTool, szResultLabel)  );
		break;
	case SG_MERGETOOL__STATUS__ERROR:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Merge tool '%s' was canceled or exited with an error.", szResultTool)  );
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Created merged value '%s' with partial result.", szResultLabel)  );
		break;
	case SG_MERGETOOL__STATUS__NOT_ATTEMPTED:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Merge tool did not attempt the merge, no value created.")  );
		break;
	case SG_MERGETOOL__STATUS__EXEC_ERROR:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "There was an error launching the merge tool, no value created.")  );
		break;
	case SG_MERGETOOL__STATUS__NO_OUTPUT:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "The merge tool did not generate any output, no value created.")  );
		break;
	default:
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Unknown merge tool result: %i", iResult)  );
		break;
	}

	// if the new value was automatically accepted as the choice's resolution
	// then let the user know
	SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );
	if (pAcceptedValue != NULL && pAcceptedValue == pResultValue)
	{
		SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "New merged value automatically accepted as its choice's resolution.")  );
	}

fail:
	return;
}

/**
 * User data used by _resolve__get_choice__index__choice.
 */
typedef struct _resolve__get_choice__index__choice__data
{
	SG_uint32           uIndex;  //< [in] Index of the choice to retrieve.
	SG_resolve__choice* pChoice; //< [out] Choice at the given index.
}
_resolve__get_choice__index__choice__data;

/**
 * SG_resolve__choice__callback that stops iterating after N calls.
 * It then returns whichever choice it stopped at.
 */
void _resolve__get_choice__index__choice(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The current choice.
	void*               pUserData, //< [in] A _resolve__get_choice__index__choice__data instance with additional parameters.
	SG_bool*            pContinue  //< [out] Whether or not iteration should continue.
	)
{
	_resolve__get_choice__index__choice__data* pData = (_resolve__get_choice__index__choice__data*)pUserData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	if (pData->uIndex == 0u)
	{
		pData->pChoice = pChoice;
		*pContinue = SG_FALSE;
	}
	else
	{
		pData->uIndex -= 1u;
		*pContinue = SG_TRUE;
	}

fail:
	return;
}

/**
 * Retrieve the choice at a specified index,
 * either in a whole resolve or a single item.
 */
void _resolve__get_choice__index(
	SG_context*          pCtx,            //< [in] [out] Error and context info.
	SG_resolve*          pResolve,        //< [in] The resolve to get a choice from.
	SG_resolve__item*    pItem,           //< [in] If non-NULL, only consider choices on this item.
	SG_uint32            uIndex,          //< [in] The index of the choice to get.
	SG_bool              bOnlyUnresolved, //< [in] Whether or not to only consider unresolved choices.
	SG_resolve__choice** pChoice          //< [out] The retrieved choice.
	)
{
	_resolve__get_choice__index__choice__data cData;

	SG_NULLARGCHECK(pCtx);
	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(pChoice);

	cData.uIndex  = uIndex;
	cData.pChoice = NULL;

	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__foreach_choice(pCtx, pResolve, bOnlyUnresolved, _resolve__get_choice__index__choice, (void*)&cData)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_resolve__item__foreach_choice(pCtx, pItem, bOnlyUnresolved, _resolve__get_choice__index__choice, (void*)&cData)  );
	}

	*pChoice = cData.pChoice;

fail:
	return;
}


/*
**
** Private Functionality - Resolve Callbacks
**
*/

/**
 * User data used by choice callbacks that list choices.
 * Currently _resolve__choice__write and _resolve__choice__interact.
 */
typedef struct _resolve__choice__list__data
{
	SG_uint32 uIndent; //< [in] Base amount to indent the display.
	SG_uint32 uMax;    //< [in] Maximum number of choices to handle.
	                   //<      If zero, all available choices will be handled.
	SG_uint32 uCount;  //< [in] Number of choices handled so far.
}
_resolve__choice__list__data;

/**
 * An SG_resolve__choice__callback that writes information about the item to stdout.
 */
static void _resolve__choice__write(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The current choice.
	void*               pUserData, //< [in] A _resolve__choice__list__data structure with additional parameters.
	SG_bool*            pContinue  //< [out] Whether or not iteration should continue.
	)
{
	_resolve__choice__list__data* pData = (_resolve__choice__list__data*)pUserData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// write the choice out
	SG_ERR_CHECK(  _resolve__write_choice(pCtx, pChoice, NULL, pData->uIndent)  );

	// if we've hit our choice limit, stop iterating
	pData->uCount += 1u;
	if (
		   pData->uMax != 0u
		&& pData->uCount >= pData->uMax
		)
	{
		*pContinue = SG_FALSE;
	}
	else
	{
		*pContinue = SG_TRUE;
	}

fail:
	return;
}


/*
**
** Public Functionality
**
*/

/**
 * Lists conflicts one by one and asks the user what to do with each.
 */
void do_cmd_resolve2__list__interactive(
	SG_context*        pCtx,    //< [in] [out] Error and context info.
	const SG_pathname* pCwd,    //< [in] The current working directory to allocate a pendingtree from.
	const char*        szPath,  //< [in] If non-NULL, only list conflicts on this item (as supplied by user).
	SG_bool            bAll,    //< [in] If true, include already resolved conflicts in the list.
	SG_uint32          uIndent, //< [in] Base amount to indent all messages.
	SG_uint32          uMax     //< [in] Maximum number of conflicts to list.
	)
{
	SG_resolve__item* pItem           = NULL;
	SG_bool           bOnlyUnresolved = SG_FALSE;
	SG_bool           bContinue       = SG_TRUE;
	SG_uint32         uChoice         = 0u;
	SG_pendingtree*   pPendingTree    = NULL;
	SG_resolve*       pResolve        = NULL;
	char*             szGid           = NULL;
	SG_bool           bWriteChoice    = SG_TRUE;
	SG_string*        sPrompt         = NULL;
	SG_bool           bAutomatic      = SG_FALSE;
	SG_string*        sInput          = NULL;
	char*             szInput         = NULL;

	SG_NULLARGCHECK(pCwd);

	// decide which choices we care about
	if (bAll == SG_FALSE)
	{
		bOnlyUnresolved = SG_TRUE;
	}
	else
	{
		bOnlyUnresolved = SG_FALSE;
	}

	// keep prompting until we decide to stop
	while (bContinue == SG_TRUE)
	{
		SG_resolve__choice* pChoice     = NULL;
		SG_bool             bMergeable  = SG_FALSE;
		const char*         szAutomatic = NULL;
		const char*         szCommand   = NULL;
		void*               pHandle     = NULL;
		SG_bool             bNext       = SG_FALSE;

		// allocate a fresh pendingtree and resolve data
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

		// if they specified a path, parse it and find the item
		// Note: It'd be nice to do this outside the loop, but we need the PendingTree.
		if (szPath != NULL)
		{
			if (szGid == NULL)
			{
				// only parse the path into a GID if we haven't done so yet
				// otherwise reuse the GID from last time, because that item
				// may no longer be at its original path from szPath
				SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, szPath, &szGid)  );
			}
			SG_ERR_CHECK(  SG_resolve__check_item__gid(pCtx, pResolve, szGid, &pItem)  );
		}

		// find the choice at the current index
		SG_ERR_CHECK(  _resolve__get_choice__index(pCtx, pResolve, pItem, uChoice, bOnlyUnresolved, &pChoice)  );
		if (pChoice == NULL)
		{
			// if we ran out of choices, then stop
			bContinue = SG_FALSE;
		}
		else
		{
			// if we need to write the current choice, do that
			if (bWriteChoice == SG_TRUE)
			{
				SG_ERR_CHECK(  _resolve__write_choice(pCtx, pChoice, NULL, uIndent)  );
				bWriteChoice = SG_FALSE;
			}

			// build the prompt
			SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sPrompt)  );
			SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
			if (bMergeable == SG_TRUE)
			{
//				SG_ERR_CHECK(  SG_string__append__sz(pCtx, sPrompt, "(v)iew (d)iff (m)erge ")  );
				SG_ERR_CHECK(  SG_string__append__sz(pCtx, sPrompt, "(d)iff (m)erge ")  );
			}
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, sPrompt, "(a)ccept (r)epeat (s)kip (q)uit (h)elp: ")  );

			// if we haven't used an automatic command on this choice yet
			// check if we should use one
			if (bAutomatic == SG_FALSE)
			{
				// check if we should automatically run a merge
				if (bMergeable == SG_TRUE)
				{
					SG_resolve__value* pBaseline = NULL;
					SG_resolve__value* pOther1   = NULL;
					SG_resolve__value* pOther2   = NULL;
					SG_resolve__value* pWorking1 = NULL;
					SG_resolve__value* pWorking2 = NULL;
					SG_resolve__value* pMerge1   = NULL;
					SG_resolve__value* pMerge2   = NULL;

					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__BASELINE, 0u, &pBaseline)  );
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__OTHER, 1u, &pOther1)  );
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__OTHER, 2u, &pOther2)  );
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__WORKING, 1u, &pWorking1)  );
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__WORKING, 2u, &pWorking2)  );
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__MERGE, 1u, &pMerge1)  );
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__MERGE, 2u, &pMerge2)  );

					// continue if we have a "baseline" value,
					// exactly one "other" and "merge" value,
					// and no more than one "working" value
					if (
						   pBaseline != NULL
						&& pOther1   != NULL
						&& pOther2   == NULL
						&& pWorking2 == NULL
						&& pMerge1   != NULL
						&& pMerge2   == NULL
						)
					{
						SG_bool bMerged        = SG_FALSE;
						SG_bool bAutomatically = SG_FALSE;

						// make sure the single merge value was automatically generated
						SG_ERR_CHECK(  SG_resolve__value__get_merge(pCtx, pMerge1, &bMerged, NULL, NULL, &bAutomatically, NULL, NULL)  );
						if (bMerged == SG_TRUE && bAutomatically == SG_TRUE)
						{
							const char* szWorking = NULL;
							const char* szMerge   = NULL;

							// compare the HIDs of the "merge" and "working" values
							// to make sure the user hasn't modified the working value since the automerge created it
							SG_ERR_CHECK(  SG_resolve__value__get_data__sz(pCtx, pWorking1, &szWorking)  );
							SG_ERR_CHECK(  SG_resolve__value__get_data__sz(pCtx, pMerge1, &szMerge)  );
							if (SG_stricmp(szWorking, szMerge) == 0)
							{
								SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sInput, "merge %s %s", SG_RESOLVE__LABEL__BASELINE, SG_RESOLVE__LABEL__OTHER)  );
								szAutomatic = "Obvious merge run automatically, close tool without saving to cancel.";
							}
						}
					}
				}
			}

			// check if we have an automatic command
			if (sInput == NULL)
			{
				// no automatic command, ask the user for one
				SG_ERR_CHECK(  _resolve__stdin(pCtx, uIndent, &sInput, "%s", SG_string__sz(sPrompt))  );
			}
			else
			{
				// automatic command, write it out with the prompt
				SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "%s%s", SG_string__sz(sPrompt), SG_string__sz(sInput))  );
				if (szAutomatic != NULL)
				{
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "%s", szAutomatic)  );
				}

				// mark that we've used an automatic command
				bAutomatic = SG_TRUE;
			}

			// unwrap the string, we just need the buffer
			SG_ERR_CHECK(  SG_string__sizzle(pCtx, &sInput, (SG_byte**)&szInput, NULL)  );

			// get the first token they entered, which will be the command
			SG_ERR_CHECK(  _SG_strtok(pCtx, szInput, NULL, SG_TRUE, &szCommand, &pHandle)  );
			if (szCommand != NULL)
			{
				const char*        szToken  = NULL;
				const char**       ppNext   = NULL;
				const char*        szValue1 = NULL;
				const char*        szValue2 = NULL;
				const char*        szTool   = NULL;
				const char*        szLabel  = NULL;
				SG_bool            bUsage   = SG_FALSE;
				SG_resolve__value* pValue1  = NULL;
				SG_resolve__value* pValue2  = NULL;

				// keep getting tokens until we run out
				SG_ERR_CHECK(  _SG_strtok(pCtx, NULL, NULL, SG_TRUE, &szToken, &pHandle)  );
				while (szToken != NULL)
				{
					if (SG_stricmp(szToken, "--tool") == 0)
					{
						if (ppNext == NULL)
						{
							// expect a tool name next
							ppNext = &szTool;
						}
						else
						{
							// usage error
							bUsage = SG_TRUE;
						}
					}
					else if (SG_stricmp(szToken, "--label") == 0)
					{
						if (ppNext == NULL)
						{
							// expect a label name next
							ppNext = &szLabel;
						}
						else
						{
							// usage error
							bUsage = SG_TRUE;
						}
					}
					else
					{
						if (ppNext == NULL)
						{
							// we're not expecting anything
							// we'll assume the token is a value name
							if (szValue1 == NULL)
							{
								szValue1 = szToken;
							}
							else if (szValue2 == NULL)
							{
								szValue2 = szToken;
							}
							else
							{
								// usage error
								bUsage = SG_TRUE;
							}
						}
						else
						{
							// we're expecting something
							// assign the token to that
							*ppNext = szToken;
							ppNext = NULL;
						}
					}

					// get the next token
					SG_ERR_CHECK(  _SG_strtok(pCtx, NULL, NULL, SG_TRUE, &szToken, &pHandle)  );
				}

				// try parsing non-NULL value names into values
				if (szValue1 != NULL)
				{
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szValue1, 0u, &pValue1)  );
				}
				if (szValue2 != NULL)
				{
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szValue2, 0u, &pValue2)  );
				}

				// decide what to do
				if (szValue1 != NULL && pValue1 == NULL)
				{
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Unknown value: %s", szValue1)  );
				}
				else if (szValue2 != NULL && pValue2 == NULL)
				{
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Unknown value: %s", szValue2)  );
				}
				else if (
					   SG_stricmp(szCommand, "accept") == 0
					|| SG_stricmp(szCommand, "a") == 0
					)
				{
					// accept
					if (
						   bUsage  == SG_TRUE
						|| pValue1 == NULL
						|| szTool  != NULL
						|| szLabel != NULL
						)
					{
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Usage: a[ccept] VALUE")  );
					}
					else
					{
						SG_resolve__value* pAcceptedValue = NULL;
						SG_bool            bOverwrite     = SG_FALSE;

						// check if the choice already has a resolution
						SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );
						if (pValue1 == pAcceptedValue)
						{
							SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "The conflict is already resolved using the specified value: %s", szValue1)  );
						}
						else if (pAcceptedValue != NULL)
						{
							const char* szAcceptedLabel = NULL;

							SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pAcceptedValue, &szAcceptedLabel)  );
							SG_ERR_CHECK(  _resolve__stdin(pCtx, uIndent, &sInput, "Choice already already resolved using value '%s', overwrite? (y/n): ", szAcceptedLabel)  );
							SG_ERR_CHECK(  _resolve__parse__bool(pCtx, SG_string__sz(sInput), &bOverwrite)  );
						}

						// run the accept
						if (pAcceptedValue == NULL || bOverwrite == SG_TRUE)
						{
							SG_ERR_CHECK(  SG_resolve__accept_value(pCtx, pValue1)  );
							SG_ERR_CHECK(  _resolve__write_accept(pCtx, uIndent, pValue1, pChoice, pItem)  );

							// advance to the next choice
							if (bOnlyUnresolved == SG_FALSE)
							{
								// This is only necessary if we're showing all choices.
								// If we're only showing unresolved choices, it will happen
								// automatically because the choice we were on is now resolved.
								uChoice += 1u;
							}

							// write the new choice out next time
							bWriteChoice = SG_TRUE;

							// mark that we're advancing a choice
							bNext = SG_TRUE;
						}
					}
				}
/*				else if (
					   bMergeable == SG_TRUE
					&& (
						   SG_stricmp(szCommand, "view") == 0
						|| SG_stricmp(szCommand, "v") == 0
						)
					)
				{
					// view
					if (
						   bUsage  == SG_TRUE
						|| pValue1 == NULL
						|| szLabel != NULL
						)
					{
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Usage: v[iew] VALUE [--tool TOOL]")  );
					}
					else
					{
						SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);
						//SG_ERR_CHECK(  SG_resolve__view_value(pCtx, pValue1, szTool)  );
					}
				}
*/				else if (
					   bMergeable == SG_TRUE
					&& (
						   SG_stricmp(szCommand, "diff") == 0
						|| SG_stricmp(szCommand, "d") == 0
						)
					)
				{
					// diff
					if (
						   bUsage  == SG_TRUE
						|| pValue1 == NULL
						|| pValue2 == NULL
						|| szLabel != NULL
						)
					{
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Usage: d[iff] VALUE1 VALUE2 [--tool TOOL]")  );
					}
					else
					{
						SG_ERR_CHECK(  SG_resolve__diff_values(pCtx, pValue1, pValue2, szTool)  );
					}
				}
				else if (
					   bMergeable == SG_TRUE
					&& (
						   SG_stricmp(szCommand, "merge") == 0
						|| SG_stricmp(szCommand, "m") == 0
						)
					)
				{
					// merge
					if (
						   bUsage  == SG_TRUE
						|| pValue1 == NULL
						|| pValue2 == NULL
						)
					{
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Usage: m[erge] VALUE1 VALUE2 [--tool TOOL] [--label LABEL]")  );
					}
					else
					{
						SG_resolve__value* pLabeledValue = NULL;
						SG_bool            bOverwrite    = SG_FALSE;

						// if they specified a label, see if it already exists
						if (szLabel != NULL)
						{
							// if it already exists, check if it's overwritable
							SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szLabel, 0u, &pLabeledValue)  );
							if (pLabeledValue != NULL)
							{
								SG_bool bOverwritable = SG_FALSE;

								// if it's overwritable, check if they want to overwrite it
								SG_ERR_CHECK(  SG_resolve__value__get_overwritable(pCtx, pLabeledValue, &bOverwritable)  );
								if (bOverwritable == SG_TRUE)
								{
									SG_ERR_CHECK(  _resolve__stdin(pCtx, uIndent, &sInput, "Label '%s' already exists, overwrite? (y/n): ", szLabel)  );
									SG_ERR_CHECK(  _resolve__parse__bool(pCtx, SG_string__sz(sInput), &bOverwrite)  );
								}
								else
								{
									SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Label '%s' already exists and cannot be overwritten.", szLabel)  );
								}
							}
						}

						// run the merge if we can
						if (pLabeledValue == NULL || bOverwrite == SG_TRUE)
						{
							SG_int32           iResult   = SG_MERGETOOL__STATUS__MUST_BE_LAST;
							SG_resolve__value* pResult   = NULL;
							SG_resolve__value* pAccepted = NULL;

							SG_ERR_CHECK(  SG_resolve__merge_values(pCtx, pValue1, pValue2, szTool, szLabel, &iResult, &pResult)  );
							SG_ERR_CHECK(  _resolve__write_merge(pCtx, 0u, pValue1, pValue2, iResult, pResult, pChoice, NULL)  );

							// if the result value was automatically accepted
							// then advance to the next choice
							if (pAccepted != NULL)
							{
								SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAccepted)  );
								if (pAccepted == pResult)
								{
									if (bOnlyUnresolved == SG_FALSE)
									{
										// This is only necessary if we're showing all choices.
										// If we're only showing unresolved choices, it will happen
										// automatically because the choice we were on is now resolved.
										uChoice += 1u;
									}

									// write the new choice out next time
									bWriteChoice = SG_TRUE;

									// mark that we're advancing a choice
									bNext = SG_TRUE;
								}
							}
						}
					}
				}
				else if (
						SG_stricmp(szCommand, "repeat") == 0
					|| SG_stricmp(szCommand, "r") == 0
					)
				{
					// write the choice out next time
					bWriteChoice = SG_TRUE;
				}
				else if (
						SG_stricmp(szCommand, "skip") == 0
					|| SG_stricmp(szCommand, "s") == 0
					)
				{
					// skip this choice
					// write the new choice out next time
					uChoice += 1u;
					bWriteChoice = SG_TRUE;

					// mark that we're advancing a choice
					bNext = SG_TRUE;
				}
				else if (
						SG_stricmp(szCommand, "quit") == 0
					|| SG_stricmp(szCommand, "q") == 0
					)
				{
					// quit prompting
					bContinue = SG_FALSE;
				}
				else if (
						SG_stricmp(szCommand, "help") == 0
					|| SG_stricmp(szCommand, "h") == 0
					)
				{
					// help or invalid command, show help/usage
					if (bMergeable == SG_TRUE)
					{
/*						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "v[iew] VALUE [--tool TOOL]")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "View a specified value in a viewer tool (not implemented).")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "The default tool for this situation my optionally be overridden.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "This command is only available for mergeable conflicts.")  );
*/						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "d[iff] VALUE1 VALUE2 [--tool TOOL]")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Diff two specified values in a diff tool.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "The default tool for this situation my optionally be overridden.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "This command is only available for mergeable conflicts.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "m[erge] VALUE1 VALUE2 [--tool TOOL] [--label LABEL]")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Merge two specified values into a new value using a merge tool.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "The default tool for this situation my optionally be overridden.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "The label for the new value may optionally be specified.")  );
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "This command is only available for mergeable conflicts.")  );
					}
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "a[ccept] VALUE")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Accept a specified value as the resolution for this conflict.")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "r[epeat]")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Display the details of the current conflict again.")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "s[kip]")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Skip this conflict without resolving it.")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "q[uit]")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Stop resolving conflicts.")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "h[elp]")  );
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent + 1u, "Display this help message.")  );
				}
				else
				{
					SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Unknown command: %s", szCommand)  );
				}
			}

			// add a blank line after whatever output we had
			SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, NULL)  );

			// save any data that changed
			SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );

			// free everything so we can reload it for the next choice
			// Note: don't free szGid, we want to reuse it next time.
			SG_RESOLVE_NULLFREE(pCtx, pResolve);
			SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
			SG_STRING_NULLFREE(pCtx, sPrompt);
			SG_STRING_NULLFREE(pCtx, sInput);
			SG_NULLFREE(pCtx, szInput);

			// if we're advancing to the next choice
			// then do a few extra things
			if (bNext == SG_TRUE)
			{
				// if we have a max choice count,
				// decrement it and stop if we hit zero
				if (uMax > 0u)
				{
					uMax -= 1u;
					if (uMax == 0u)
					{
						bContinue = SG_FALSE;
					}
				}

				// reset choice-based data
				bAutomatic = SG_FALSE;

				// we're done with choice-advancing tasks
				bNext = SG_FALSE;
			}
		}
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	SG_STRING_NULLFREE(pCtx, sPrompt);
	SG_STRING_NULLFREE(pCtx, sInput);
	SG_NULLFREE(pCtx, szInput);
	return;
}

/**
 * Lists conflicts.
 */
void do_cmd_resolve2__list__write(
	SG_context*        pCtx,    //< [in] [out] Error and context info.
	const SG_pathname* pCwd,    //< [in] The current working directory to allocate a pendingtree from.
	const char*        szPath,  //< [in] If non-NULL, only list conflicts on this item (as supplied by user).
	SG_bool            bAll,    //< [in] If true, include already resolved conflicts in the list.
	SG_uint32          uIndent, //< [in] Base amount to indent all messages.
	SG_uint32          uMax     //< [in] Maximum number of conflicts to list.
	)
{
	SG_pendingtree*              pPendingTree    = NULL;
	SG_resolve*                  pResolve        = NULL;
	SG_bool                      bOnlyUnresolved = SG_FALSE;
	char*                        szGid           = NULL;
	_resolve__choice__list__data cData;

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// setup the callback's data
	cData.uIndent = uIndent;
	cData.uMax    = uMax;
	cData.uCount  = 0u;

	// decide what we're listing
	if (bAll == SG_FALSE)
	{
		bOnlyUnresolved = SG_TRUE;
	}
	else
	{
		bOnlyUnresolved = SG_FALSE;
	}

	// decide which set of choices to iterate
	if (szPath == NULL)
	{
		// no item specified, iterate all of them
		SG_ERR_CHECK(  SG_resolve__foreach_choice(pCtx, pResolve, bOnlyUnresolved, _resolve__choice__write, (void*)&cData)  );
	}
	else
	{
		SG_resolve__item* pItem = NULL;

		// an item was specified, just iterate that one
		SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, szPath, &szGid)  );
		SG_ERR_CHECK(  SG_resolve__get_item__gid(pCtx, pResolve, szGid, &pItem)  );
		SG_ERR_CHECK(  SG_resolve__item__foreach_choice(pCtx, pItem, bOnlyUnresolved, _resolve__choice__write, (void*)&cData)  );
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Accepts a value on multiple conflicts.
 */
void do_cmd_resolve2__accept__all(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	const SG_pathname* pCwd,       //< [in] The current working directory to allocate a pendingtree from.
	const char*        szValue,    //< [in] Name of the value to accept.
	const char*        szArg1,     //< [in] Either the path to a file or the name of a conflict.
	                               //<      If NULL, the accept will not be restricted by file or conflict type.
	const char*        szArg2,     //< [in] If szArg1 is a path, this may be the name of a conflict.
	                               //<      If NULL, the accept will only be restricted by szArg1.
	SG_bool            bOverwrite, //< [in] If true, overwrite the resolutions of matching conflicts that are already resolved.
	                               //<      If false, ignore matching conflicts that are already resolved.
	SG_uint32          uIndent     //< [in] Base amount to indent all messages.
	)
{
	SG_bool         bContinue    = SG_TRUE;
	SG_uint32       uChoice      = 0u;
	SG_pendingtree* pPendingTree = NULL;
	SG_resolve*     pResolve     = NULL;
	char*           szGid        = NULL;

	// keep going until we're done
	while (bContinue == SG_TRUE)
	{
		SG_resolve__item*   pItem           = NULL;
		SG_resolve__state   eState          = SG_RESOLVE__STATE__COUNT;
		SG_bool             bOnlyUnresolved = SG_FALSE;
		SG_resolve__choice* pChoice         = NULL;

		// allocate a fresh pendingtree and resolve data
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

		// parse the arguments they specified
		// Note: It'd be nice to do this outside the loop, but we need the PendingTree.
		if (szArg1 != NULL)
		{
			// szArg1 might be a conflict or a path
			// try parsing a conflict first
			SG_ERR_TRY(  SG_resolve__state__parse_name(pCtx, szArg1, &eState)  );
			SG_ERR_CATCH_IGNORE(SG_ERR_NOT_FOUND)
			SG_ERR_CATCH_END;
			if (eState == SG_RESOLVE__STATE__COUNT)
			{
				// not a conflict, try parsing a path
				if (szGid == NULL)
				{
					// only parse the path into a GID if we haven't done so yet
					// otherwise reuse the GID from last time, because that item
					// may no longer be at its original path from szArg1
					SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, szArg1, &szGid)  );
				}
				SG_ERR_CHECK(  SG_resolve__get_item__gid(pCtx, pResolve, szGid, &pItem)  );

				// try parsing szArg2 into a conflict instead
				if (szArg2 != NULL)
				{
					SG_bool bFailed = SG_FALSE;

					SG_ERR_TRY(  SG_resolve__state__parse_name(pCtx, szArg2, &eState)  );
					SG_ERR_CATCH_SET(SG_ERR_NOT_FOUND, bFailed, SG_TRUE)
					SG_ERR_CATCH_END;

					if (bFailed == SG_TRUE)
					{
						SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Unknown conflict type: %s", szArg2)  );
						break;
					}
				}
			}
		}

		// check if we want resolved items too
		if (bOverwrite == SG_FALSE)
		{
			bOnlyUnresolved = SG_TRUE;
		}
		else
		{
			bOnlyUnresolved = SG_FALSE;
		}

		// find the choice at the current index
		SG_ERR_CHECK(  _resolve__get_choice__index(pCtx, pResolve, pItem, uChoice, bOnlyUnresolved, &pChoice)  );
		if (pChoice == NULL)
		{
			// if we ran out of choices, then we're done
			bContinue = SG_FALSE;
		}
		else
		{
			SG_resolve__state eCurrentState = SG_RESOLVE__STATE__COUNT;

			// if we only care about one type of choice,
			// then check if this one matches
			SG_ERR_CHECK(  SG_resolve__choice__get_state(pCtx, pChoice, &eCurrentState)  );
			if (
				   eState != SG_RESOLVE__STATE__COUNT
				&& eState != eCurrentState
				)
			{
				// skip this choice
				uChoice += 1u;
			}
			else
			{
				SG_resolve__value* pValue = NULL;

				// check if this choice has a value with the specified label
				SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szValue, 0u, &pValue)  );
				if (pValue == NULL)
				{
					// skip this choice
					uChoice += 1u;
				}
				else
				{
					// accept the value
					SG_ERR_CHECK(  SG_resolve__accept_value(pCtx, pValue)  );
					SG_ERR_CHECK(  _resolve__write_accept(pCtx, uIndent, pValue, pChoice, pItem)  );

					// advance to the next choice
					if (bOnlyUnresolved == SG_FALSE)
					{
						// This is only necessary if we're iterating all choices.
						// If we're only iterating unresolved choices, it will happen
						// automatically because the choice we were on is now resolved.
						uChoice += 1u;
					}
				}
			}
		}

		// save any data that changed
		SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );

		// free everything so we can reload it for the next choice
		// Note: don't free szGid, we want to reuse it next time.
		SG_RESOLVE_NULLFREE(pCtx, pResolve);
		SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Accepts a value on a conflict.
 */
void do_cmd_resolve2__accept__single(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	const SG_pathname* pCwd,       //< [in] The current working directory to allocate a pendingtree from.
	const char*        szValue,    //< [in] Name of the value to accept.
	const char*        szPath,     //< [in] Path of the file to accept a value on, as specified by the user.
	const char*        szConflict, //< [in] The name of the conflict/state to accept a value on.
	                               //<      May be NULL if the specified item only has a single conflict.
	SG_bool            bOverwrite, //< [in] If true, the item's existing resolution may be overwritten.
	                               //<      If false, an already resolved item will produce an error.
	SG_uint32          uIndent     //< [in] Base amount to indent all messages.
	)
{
	SG_pendingtree*     pPendingTree = NULL;
	SG_resolve*         pResolve     = NULL;
	char*               szGid        = NULL;
	SG_resolve__state   eState       = SG_RESOLVE__STATE__COUNT;
	SG_resolve__item*   pItem        = NULL;
	SG_resolve__choice* pChoice      = NULL;

	SG_NULLARGCHECK(pCwd);
	SG_NULLARGCHECK(szValue);
	SG_NULLARGCHECK(szPath);

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// parse the specified path
	SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, szPath, &szGid)  );

	// if they specified a conflict, parse that
	if (szConflict != NULL)
	{
		SG_ERR_CHECK(  SG_resolve__state__parse_name(pCtx, szConflict, &eState)  );
	}

	// find a single item/choice to accept the value on
	SG_ERR_CHECK(  _resolve__find_choice(pCtx, 0u, pResolve, szGid, eState, &pItem, &pChoice)  );
	if (pItem != NULL && pChoice != NULL)
	{
		SG_resolve__value* pAcceptedValue  = NULL;
		const char*        szAcceptedLabel = NULL;

		// get the choice's currently accepted value
		SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );
		if (pAcceptedValue != NULL)
		{
			SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pAcceptedValue, &szAcceptedLabel)  );
		}

		// check for problems
		if (pAcceptedValue != NULL && bOverwrite == SG_FALSE)
		{
			SG_ERR_CHECK(  _resolve__stdout(pCtx, 0u, "The conflict was already resolved by accepting value: %s", szAcceptedLabel)  );
			SG_ERR_CHECK(  _resolve__stdout(pCtx, 0u, "Use --overwrite to accept a different value.")  );
		}
		else if (SG_stricmp__null(szAcceptedLabel, szValue) == 0)
		{
			SG_ERR_CHECK(  _resolve__stdout(pCtx, 0u, "The conflict is already resolved using the specified value: %s", szValue)  );
		}
		else
		{
			// do the work
			SG_resolve__value* pValue = NULL;

			SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, szValue, 0u, &pValue)  );
			SG_ERR_CHECK(  SG_resolve__accept_value(pCtx, pValue)  );
			SG_ERR_CHECK(  _resolve__write_accept(pCtx, uIndent, pValue, pChoice, pItem)  );
		}
	}

	// save any data that changed
	SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Entry point for resolve command implementation.
 */
void do_cmd_resolve2(
	SG_context*      pCtx,     //< [in] [out] Error and context info.
	SG_option_state* pOptions, //< [in] Parsed option state.
	SG_uint32        uArgs,    //< [in] Number of unparsed arguments.
	const char**     ppArgs    //< [in] Unparsed arguments.
	)
{
	SG_uint32       uIndent      = 0u;
	SG_pathname*    pCwd         = NULL;
	SG_pendingtree* pPendingTree = NULL;
	SG_resolve*     pResolve     = NULL;
	char*           szGid        = NULL;
	SG_pathname*    pFilename    = NULL;

	SG_NULLARGCHECK(pOptions);
	SG_ARGCHECK(ppArgs != NULL || uArgs == 0u, ppArgs);

	// get the current working directory
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pCwd)  );

	// do a complete pendingtree scan
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__scan(pCtx, pPendingTree, SG_TRUE, SG_TRUE, SG_TRUE, gpBlankFilespec)  );
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

	// decide what we're doing
	if (uArgs == 0)
	{
		// default subcommand is: list --interactive
		SG_ERR_CHECK(  do_cmd_resolve2__list__interactive(pCtx, pCwd, NULL, SG_FALSE, uIndent, 0u)  );
	}
	else if (SG_stricmp(ppArgs[0], "list") == 0)
	{
		// call the appropriate function
		if (pOptions->bInteractive == SG_TRUE)
		{
			SG_ERR_CHECK(  do_cmd_resolve2__list__interactive(pCtx, pCwd, uArgs > 1u ? ppArgs[1] : NULL, pOptions->bAll, uIndent, 0u)  );
		}
		else
		{
			SG_ERR_CHECK(  do_cmd_resolve2__list__write(pCtx, pCwd, uArgs > 1u ? ppArgs[1] : NULL, pOptions->bAll, uIndent, 0u)  );
		}
	}
	else if (SG_stricmp(ppArgs[0], "next") == 0)
	{
		// call the appropriate function
		if (pOptions->bInteractive == SG_TRUE)
		{
			SG_ERR_CHECK(  do_cmd_resolve2__list__interactive(pCtx, pCwd, uArgs > 1u ? ppArgs[1] : NULL, pOptions->bAll, uIndent, 1u)  );
		}
		else
		{
			SG_ERR_CHECK(  do_cmd_resolve2__list__write(pCtx, pCwd, uArgs > 1u ? ppArgs[1] : NULL, pOptions->bAll, uIndent, 1u)  );
		}
	}
	else if (SG_stricmp(ppArgs[0], "accept") == 0)
	{
		const char* szValue = NULL;

		// parse the arguments
		if (uArgs < 2u)
		{
			SG_ERR_THROW(SG_ERR_USAGE);
		}
		szValue = ppArgs[1];

		// check if we're accepting for a single choice or many
		if (pOptions->bAll == SG_FALSE)
		{
			// make sure they specified an item
			if (uArgs < 3u)
			{
				SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "To resolve one conflict specify the file (and conflict if it has several).")  );
				SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "To resolve multiple conflicts use --all.")  );
			}
			else
			{
				SG_ERR_CHECK(  do_cmd_resolve2__accept__single(pCtx, pCwd, ppArgs[1], ppArgs[2], uArgs > 3u ? ppArgs[3] : NULL, pOptions->bOverwrite, uIndent)  );
			}
		}
		else
		{
			const char* szArg1 = NULL;
			const char* szArg2 = NULL;

			// parse the additional arguments
			if (uArgs > 2u)
			{
				szArg1 = ppArgs[2];
			}
			if (uArgs > 3u)
			{
				szArg2 = ppArgs[3];
			}

			SG_ERR_CHECK(  do_cmd_resolve2__accept__all(pCtx, pCwd, szValue, szArg1, szArg2, pOptions->bOverwrite, uIndent)  );
		}
	}
/*	else if (SG_stricmp(ppArgs[0], "view") == 0)
	{
		SG_resolve__state   eState     = SG_RESOLVE__STATE__COUNT;
		SG_resolve__choice* pChoice    = NULL;
		SG_bool             bMergeable = SG_FALSE;
		SG_resolve__value*  pValue     = NULL;

		if (uArgs < 3)
		{
			SG_ERR_THROW(SG_ERR_USAGE);
		}
		if (uArgs > 3u)
		{
			SG_ERR_CHECK(  SG_resolve__state__parse_name(pCtx, ppArgs[3], &eState)  );
		}

		// load the pendingtree and resolve data
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

		// find the requested choice
		SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, ppArgs[2], &szGid)  );
		SG_ERR_CHECK(  _resolve__find_choice(pCtx, uIndent, pResolve, szGid, eState, NULL, &pChoice)  );
		if (pChoice != NULL)
		{
			// make sure this is a mergeable choice
			SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
			if (bMergeable == SG_FALSE)
			{
				SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Cannot view values on non-mergeable conflicts.")  );
			}
			else
			{
				SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, ppArgs[1], 0u, &pValue)  );
				SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);
				//SG_ERR_CHECK(  SG_resolve__view_value(pCtx, pValue, pOptions->psz_tool)  );
			}
		}

		// save the resolve/pendingtree data back out
		SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );
	}
*/	else if (SG_stricmp(ppArgs[0], "diff") == 0)
	{
		SG_resolve__state   eState  = SG_RESOLVE__STATE__COUNT;
		SG_resolve__choice* pChoice = NULL;

		if (uArgs < 4)
		{
			SG_ERR_THROW(SG_ERR_USAGE);
		}
		if (uArgs > 4u)
		{
			SG_ERR_CHECK(  SG_resolve__state__parse_name(pCtx, ppArgs[4], &eState)  );
		}

		// load the pendingtree and resolve data
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

		// find the requested choice
		SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, ppArgs[3], &szGid)  );
		SG_ERR_CHECK(  _resolve__find_choice(pCtx, uIndent, pResolve, szGid, eState, NULL, &pChoice)  );
		if (pChoice != NULL)
		{
			SG_bool bMergeable = SG_FALSE;

			// make sure this is a mergeable choice
			SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
			if (bMergeable == SG_FALSE)
			{
				SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Cannot diff values on non-mergeable conflicts.")  );
			}
			else
			{
				SG_resolve__value* pLeftValue  = NULL;
				SG_resolve__value* pRightValue = NULL;

				SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, ppArgs[1], 0u, &pLeftValue)  );
				SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, ppArgs[2], 0u, &pRightValue)  );
				SG_ERR_CHECK(  SG_resolve__diff_values(pCtx, pLeftValue, pRightValue, pOptions->psz_tool)  );
			}
		}

		// save the resolve/pendingtree data back out
		SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );
	}
	else if (SG_stricmp(ppArgs[0], "merge") == 0)
	{
		SG_resolve__state   eState  = SG_RESOLVE__STATE__COUNT;
		SG_resolve__item*   pItem   = NULL;
		SG_resolve__choice* pChoice = NULL;

		if (uArgs < 4)
		{
			SG_ERR_THROW(SG_ERR_USAGE);
		}
		if (uArgs > 4u)
		{
			SG_ERR_CHECK(  SG_resolve__state__parse_name(pCtx, ppArgs[4], &eState)  );
		}

		// load the pendingtree and resolve data
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

		// find the requested choice
		SG_ERR_CHECK(  _resolve__parse__path(pCtx, pPendingTree, ppArgs[3], &szGid)  );
		SG_ERR_CHECK(  _resolve__find_choice(pCtx, uIndent, pResolve, szGid, eState, &pItem, &pChoice)  );
		if (pItem != NULL && pChoice != NULL)
		{
			SG_bool bMergeable = SG_FALSE;

			// make sure this is a mergeable choice
			SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
			if (bMergeable == SG_FALSE)
			{
				SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Cannot merge values on non-mergeable conflicts.")  );
			}
			else
			{
				SG_resolve__value* pBaselineValue = NULL;
				SG_resolve__value* pOtherValue    = NULL;
				SG_resolve__value* pExistingValue = NULL;
				SG_bool            bOverwrite     = pOptions->bOverwrite;

				// get the values
				SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, ppArgs[1], 0u, &pBaselineValue)  );
				SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, ppArgs[2], 0u, &pOtherValue)  );

				// if they specified a label, check if it's already in use
				if (pOptions->psz_label != NULL)
				{
					SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, pOptions->psz_label, 0u, &pExistingValue)  );
					if (pExistingValue != NULL)
					{
						SG_bool bOverwritable = SG_FALSE;

						// check if the existing value can be overwritten
						SG_ERR_CHECK(  SG_resolve__value__get_overwritable(pCtx, pExistingValue, &bOverwritable)  );
						if (bOverwritable == SG_FALSE)
						{
							// can't overwrite this one
							SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "Cannot overwrite automatic value: %s", pOptions->psz_label)  );
							bOverwrite = SG_FALSE;
						}
						else if (bOverwrite == SG_FALSE)
						{
							// could overwrite this one, but they didn't supply that option
							SG_ERR_CHECK(  _resolve__stdout(pCtx, uIndent, "A value with label '%s' already exists, use --overwrite to replace it.", pOptions->psz_label)  );
						}
					}
				}

				// if we can run the merge, do it
				if (pExistingValue == NULL || bOverwrite == SG_TRUE)
				{
					SG_int32           iResult      = SG_MERGETOOL__STATUS__MUST_BE_LAST;
					SG_resolve__value* pResultValue = NULL;

					SG_ERR_CHECK(  SG_resolve__merge_values(pCtx, pBaselineValue, pOtherValue, pOptions->psz_tool, pOptions->psz_label, &iResult, &pResultValue)  );
					SG_ERR_CHECK(  _resolve__write_merge(pCtx, uIndent, pBaselineValue, pOtherValue, iResult, pResultValue, pChoice, pItem)  );
				}
			}
		}

		// save the resolve/pendingtree data back out
		SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );
	}
	else
	{
		SG_bool bExists = SG_FALSE;

		// check if the first argument is an existing file
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pFilename, ppArgs[0])  );
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pFilename, &bExists, NULL, NULL)  );
		if (bExists == SG_FALSE)
		{
			SG_ERR_THROW(SG_ERR_USAGE);
		}
		else
		{
			SG_ERR_CHECK(  do_cmd_resolve2__list__interactive(pCtx, pCwd, ppArgs[0], pOptions->bAll, uIndent, 0u)  );
		}
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pCwd);
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	SG_PATHNAME_NULLFREE(pCtx, pFilename);
	return;
}
