/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "sg_fs3__private.h"

/* TODO handle SQLITE_BUSY */

#define MY_SLEEP_MS      100
#define MY_TIMEOUT_MS  30000

struct _SG_treendx
{
	SG_repo* pRepo;
    SG_uint64 iDagNum;

	SG_pathname* pPath_db;  /* the sqlite db file */

    sqlite3* psql;
	SG_bool			bInTransaction;		// TODO sqlite3 does not allow nested transactions, right?

    // TODO bQueryOnly isn't used
    SG_bool bQueryOnly;

    struct
    {
        SG_bool b_in_update;
        SG_uint32 count_so_far;
        SG_uint32 count_changesets;
        sqlite3_stmt* pStmt_changes;
    } update;
};

static void sg_treendx__create_indexes(SG_context* pCtx, SG_treendx* pdbc)
{
    SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pdbc->psql, "CREATE INDEX IF NOT EXISTS changes_gid ON changes (gid)")  );

    // TODO create other indexes on the changes table?

fail:
    ;
}

static void sg_treendx__create_db(SG_context* pCtx, SG_treendx* pTreeNdx)
{
	SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pTreeNdx->psql, "CREATE TABLE changes (gid VARCHAR NOT NULL, csid VARCHAR NOT NULL, parent VARCHAR NOT NULL, generation INTEGER NOT NULL, hid VARCHAR NULL, dir VARCHAR NULL, xattrs VARCHAR NULL, name VARCHAR NULL, bits INTEGER NULL, type INTEGER NULL)")  );

	return;
fail:
	return;
}

void SG_treendx__open(
	SG_context* pCtx,
	SG_repo* pRepo,
	SG_uint64 iDagNum,
	SG_pathname* pPath,
	SG_bool bQueryOnly,
	SG_treendx** ppNew
	)
{
	SG_treendx* pTreeNdx = NULL;
    SG_bool b_exists = SG_FALSE;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pTreeNdx)  );

    pTreeNdx->pRepo = pRepo;
    pTreeNdx->pPath_db = pPath;
    pPath = NULL;
    pTreeNdx->iDagNum = iDagNum;
    pTreeNdx->bQueryOnly = bQueryOnly;
	pTreeNdx->bInTransaction = SG_FALSE;

    SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pTreeNdx->pPath_db, &b_exists, NULL, NULL)  );

    if (b_exists)
    {
        SG_ERR_CHECK(  sg_sqlite__open__pathname(pCtx, pTreeNdx->pPath_db, &pTreeNdx->psql)  );
    }
    else
    {
        SG_ERR_CHECK(  sg_sqlite__create__pathname(pCtx, pTreeNdx->pPath_db,&pTreeNdx->psql)  );
        SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pTreeNdx->psql, "PRAGMA journal_mode=WAL")  );

        SG_ERR_CHECK(  sg_treendx__create_db(pCtx, pTreeNdx)  );
    }

	*ppNew = pTreeNdx;

	return;
fail:
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    SG_TREENDX_NULLFREE(pCtx, pTreeNdx);
}

void SG_treendx__begin_update_multiple(
        SG_context* pCtx, 
        SG_treendx* pdbc, 
        SG_uint32 count_changesets
        )
{
    SG_ASSERT(!pdbc->update.b_in_update);

    memset(&pdbc->update, 0, sizeof(pdbc->update));

    pdbc->update.b_in_update = SG_TRUE;
    pdbc->update.count_changesets = count_changesets;

    SG_ERR_CHECK(  sg_sqlite__exec__retry(pCtx, pdbc->psql, "BEGIN TRANSACTION", MY_SLEEP_MS, MY_TIMEOUT_MS)  );
	pdbc->bInTransaction = SG_TRUE;

    SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pdbc->psql, &pdbc->update.pStmt_changes, "INSERT INTO changes (gid, csid, parent, generation, hid, dir, xattrs, name, bits, type) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)")  );

fail:
    ;
}

void SG_treendx__end_update_multiple(SG_context* pCtx, SG_treendx* pdbc)
{
    SG_ASSERT(pdbc->update.b_in_update);

    SG_ASSERT(pdbc->update.count_changesets == pdbc->update.count_so_far);

	SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pdbc->update.pStmt_changes)  );

    SG_ERR_CHECK(  sg_treendx__create_indexes(pCtx, pdbc)  );

    SG_ERR_CHECK(  sg_sqlite__exec__retry(pCtx, pdbc->psql, "COMMIT TRANSACTION", MY_SLEEP_MS, MY_TIMEOUT_MS)  );
	pdbc->bInTransaction = SG_FALSE;

    pdbc->update.b_in_update = SG_FALSE;

fail:
    ;
}

void SG_treendx__abort_update_multiple(SG_context* pCtx, SG_treendx* pdbc)
{
    SG_ASSERT(pdbc->update.b_in_update);

	SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pdbc->update.pStmt_changes)  );

    SG_ERR_CHECK(  sg_sqlite__exec__retry(pCtx, pdbc->psql, "ROLLBACK TRANSACTION", MY_SLEEP_MS, MY_TIMEOUT_MS)  );
	pdbc->bInTransaction = SG_FALSE;

    pdbc->update.b_in_update = SG_FALSE;

fail:
    ;
}

void SG_treendx__update_one_changeset(
        SG_context* pCtx,
        SG_treendx* pTreeNdx,
        const char* psz_csid,
        SG_changeset* pcs
        )
{
    SG_vhash* pvh_changes_by_parent = NULL;
    SG_int32 generation = -1;

	SG_NULLARGCHECK_RETURN(psz_csid);
	SG_NULLARGCHECK_RETURN(pcs);
	SG_NULLARGCHECK_RETURN(pTreeNdx);

    SG_ERR_CHECK(  SG_changeset__get_generation(pCtx, pcs, &generation)  );

    SG_ERR_CHECK(  SG_changeset__tree__get_changes(pCtx, pcs, &pvh_changes_by_parent)  );
    if (pvh_changes_by_parent)
    {
        SG_uint32 count_parents = 0;
        SG_uint32 i_parent = 0;

        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_changes_by_parent, &count_parents)  );

        for (i_parent=0; i_parent<count_parents; i_parent++)
        {
            SG_uint32 count_changes = 0;
            SG_uint32 i_chg = 0;
            const char* psz_csid_parent = NULL;
            SG_vhash* pvh_changes = NULL;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_changes_by_parent, i_parent, &psz_csid_parent, &pvh_changes)  );
            SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_changes, &count_changes)  );

            for (i_chg=0; i_chg<count_changes; i_chg++)
            {
                const SG_variant* pv = NULL;
                const char* psz_gid = NULL;
                SG_vhash* pvh_info = NULL;
                const char* psz_hid = NULL;
                const char* psz_hid_xattrs = NULL;
                const char* psz_dir = NULL;
                const char* psz_name = NULL;
                SG_int64 type = 0;
                SG_int64 bits = 0;
                SG_bool b_type = SG_FALSE;
                SG_bool b_bits = SG_FALSE;

                SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_changes, i_chg, &psz_gid, &pv)  );
                if (SG_VARIANT_TYPE_VHASH == pv->type)
                {
                    SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pv, &pvh_info)  );
                    SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__HID, &psz_hid)  );
                    SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__NAME, &psz_name)  );
                    SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__TYPE, &b_type)  );
                    if (b_type)
                    {
                        SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__TYPE, &type)  );
                    }
                    SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__BITS, &b_bits)  );
                    if (b_bits)
                    {
                        SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__BITS, &bits)  );
                    }
                    SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__XATTRS, &psz_hid_xattrs)  );
                    SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_info, SG_CHANGEESET__TREE__CHANGES__DIR, &psz_dir)  );
                }
                else
                {
                    SG_ASSERT(SG_VARIANT_TYPE_NULL == pv->type);
                }

                SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pTreeNdx->update.pStmt_changes)  );
                SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pTreeNdx->update.pStmt_changes)  );
                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 1, psz_gid)  );
                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 2, psz_csid)  );
                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 3, psz_csid_parent)  );
                SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pTreeNdx->update.pStmt_changes, 4, generation)  );
                if (psz_hid)
                {
                    SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 5, psz_hid)  );
                }
                if (psz_dir)
                {
                    SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 6, psz_dir)  );
                }
                if (psz_hid_xattrs)
                {
                    SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 7, psz_hid_xattrs)  );
                }
                if (psz_name)
                {
                    SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pTreeNdx->update.pStmt_changes, 8, psz_name)  );
                }
                if (b_type)
                {
                    SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pTreeNdx->update.pStmt_changes, 9, (SG_int32) type)  );
                }
                if (b_bits)
                {
                    SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pTreeNdx->update.pStmt_changes, 10, (SG_int32) bits)  );
                }

                SG_ERR_CHECK(  sg_sqlite__step(pCtx, pTreeNdx->update.pStmt_changes, SQLITE_DONE)  );
            }
        }
    }

    pTreeNdx->update.count_so_far++;

fail:
    ;
}

static void sg_treendx__get_all_gidpaths(
        SG_context* pCtx, 
        SG_treendx* pndx, 
        const char* psz_gid, 
        SG_vhash* pvh_ancestors,
        SG_varray** ppResults
        )
{
	sqlite3_stmt* pStmt = NULL;
	SG_varray* pva_results = NULL;
    SG_varray* pva_paths = NULL;
	int rc = 0;
    SG_vhash* pvh_dirs = NULL;
    SG_string* pstr = NULL;
    SG_uint32 count_dirs = 0;
    SG_uint32 i_dir = 0;
    SG_vhash* pvh_new_ancestors = NULL;

	SG_ERR_CHECK_RETURN(  SG_gid__argcheck(pCtx, psz_gid)  );

    SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_dirs)  );
	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pndx->psql, &pStmt, "SELECT dir FROM changes WHERE gid='%s'", psz_gid)  );
	while ((rc = sqlite3_step(pStmt)) == SQLITE_ROW)
	{
		const char* psz_dir = (const char*) sqlite3_column_text(pStmt, 0);

        if (psz_dir && *psz_dir)
        {
            SG_bool b_ok = SG_TRUE;

            if (pvh_ancestors)
            {
                SG_bool b_anc = SG_FALSE;

                SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_ancestors, psz_dir, &b_anc)  );
                if (b_anc)
                {
                    b_ok = SG_FALSE;
                }
            }

            if (b_ok)
            {
                SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh_dirs, psz_dir)   );
            }
        }
	}
	if (rc != SQLITE_DONE)
	{
		SG_ERR_THROW(SG_ERR_SQLITE(rc));
	}

	SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pva_results)  );

    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_dirs, &count_dirs)  );

    if ( (0 == count_dirs))
    {
        SG_vhash* pvh_gidpath = NULL;

        SG_ERR_CHECK(  SG_varray__appendnew__vhash(pCtx, pva_results, &pvh_gidpath)  );
        SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvh_gidpath, "@")  );
    }
    else
    {
        if (pvh_ancestors)
        {
            SG_ERR_CHECK(  SG_vhash__alloc__copy(pCtx, &pvh_new_ancestors, pvh_ancestors)  );
        }
        else
        {
            SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_new_ancestors)  );
        }

        SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvh_new_ancestors, psz_gid)  );

        SG_ERR_CHECK(  SG_string__alloc(pCtx, &pstr)  );
        for (i_dir=0; i_dir<count_dirs; i_dir++)
        {
            const char* psz_dir_gid = NULL;
            SG_uint32 count_paths = 0;
            SG_uint32 i_path = 0;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_dirs, i_dir, &psz_dir_gid, NULL)  );
            SG_ERR_CHECK(  sg_treendx__get_all_gidpaths(pCtx, pndx, psz_dir_gid, pvh_new_ancestors, &pva_paths)  );
            SG_ERR_CHECK(  SG_varray__count(pCtx, pva_paths, &count_paths)  );
            for (i_path=0; i_path<count_paths; i_path++)
            {
                SG_vhash* pvh_gidpath = NULL;
                SG_vhash* pvh_gidpath_copy = NULL;

                SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_paths, i_path, &pvh_gidpath)  );
                SG_ERR_CHECK(  SG_varray__appendcopy__vhash(pCtx, pva_results, pvh_gidpath, &pvh_gidpath_copy)  );
                SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvh_gidpath_copy, psz_gid)  );
            }

            SG_VARRAY_NULLFREE(pCtx, pva_paths);
        }
    }

	*ppResults = pva_results;
    pva_results = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_new_ancestors);
    SG_STRING_NULLFREE(pCtx, pstr);
    SG_VARRAY_NULLFREE(pCtx, pva_results);
    SG_VARRAY_NULLFREE(pCtx, pva_paths);
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
    SG_VHASH_NULLFREE(pCtx, pvh_dirs);
}

static void sg_treendx__find_gidpath(
        SG_context* pCtx, 
        SG_repo* pRepo, 
        const char* psz_hid_treenode,
        SG_vhash* pvh_gidpath,
        SG_string** ppstr_path,
        SG_treenode_entry** pptne
        )
{
	SG_uint32 count = 0;
	SG_uint32 i = 0;
    SG_treenode* ptn_cur = NULL;
    SG_string* pstr_path = NULL;
    const SG_treenode_entry* ptne = NULL;
    char* psz_hid_treenode_cur = NULL;

    SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_hid_treenode, &psz_hid_treenode_cur)  );

    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_gidpath, &count)  );

    SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr_path)  );
	for (i=0; i<count; i++)
    {
        const char* psz_gid = NULL;
        const char* psz_hid_next = NULL;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_gidpath, i, &psz_gid, NULL)  );

        SG_TREENODE_NULLFREE(pCtx, ptn_cur);
        ptne = NULL;
        SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, psz_hid_treenode_cur, &ptn_cur)  );
        SG_NULLFREE(pCtx, psz_hid_treenode_cur);

        if (0 == strcmp(psz_gid, "@"))
        {
            SG_ERR_CHECK(  SG_string__set__sz(pCtx, pstr_path, "@")  );
            SG_ERR_CHECK(  SG_treenode__get_nth_treenode_entry__ref(pCtx, ptn_cur, 0, NULL, &ptne)  );
            SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptne, &psz_hid_next)  );
            SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_hid_next, &psz_hid_treenode_cur)  );
        }
        else
        {
            const char* psz_name = NULL;

            SG_ERR_CHECK(  SG_treenode__find_entry_by_gid(pCtx, ptn_cur, psz_gid, &ptne)  );
            if (!ptne)
            {
                SG_STRING_NULLFREE(pCtx, pstr_path);

                goto done;
            }
            SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx, ptne, &psz_name)  );
			SG_ERR_CHECK(  SG_repopath__append_entryname(pCtx, pstr_path, psz_name, SG_FALSE)  );
            SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptne, &psz_hid_next)  );
            SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_hid_next, &psz_hid_treenode_cur)  );
        }
    }

done:
    *ppstr_path = pstr_path;
    pstr_path = NULL;

    if (ptne)
    {
        SG_ERR_CHECK(  SG_treenode_entry__alloc__copy(pCtx, pptne, ptne)  );
    }

fail:
    SG_STRING_NULLFREE(pCtx, pstr_path);
    SG_NULLFREE(pCtx, psz_hid_treenode_cur);
    SG_TREENODE_NULLFREE(pCtx, ptn_cur);
}

void SG_treendx__get_path_in_dagnode(
    SG_context* pCtx, 
    SG_treendx* pndx, 
    const char* psz_gid, 
    const char* psz_csid, 
    SG_string** ppPathOfFoundObject, 
    SG_treenode_entry ** ppTreeNodeEntry
    )
{
	SG_varray* pva_gidpaths = NULL;
	char* psz_hid_treenode = NULL;
	SG_uint32 i = 0;
	SG_uint32 count = 0;
    SG_string* pstr_path = NULL;
    SG_treenode_entry* pentry = NULL;

	SG_ERR_CHECK_RETURN(  SG_gid__argcheck(pCtx, psz_gid)  );

	SG_ERR_CHECK(  sg_treendx__get_all_gidpaths(pCtx, pndx, psz_gid, NULL, &pva_gidpaths)   );
	SG_ERR_CHECK(  SG_changeset__tree__find_root_hid(pCtx, pndx->pRepo, psz_csid, &psz_hid_treenode)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_gidpaths, &count )  );

	for (i = 0; i < count; i++)
	{
        SG_vhash* pvh_gidpath = NULL;

		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_gidpaths, i, &pvh_gidpath)  );
		SG_ERR_CHECK(  sg_treendx__find_gidpath(pCtx, pndx->pRepo, psz_hid_treenode, pvh_gidpath, &pstr_path, &pentry)  );
		if (pentry)
		{
			break;
		}
        SG_STRING_NULLFREE(pCtx, pstr_path);
        SG_TREENODE_ENTRY_NULLFREE(pCtx, pentry);
	}

    if (pstr_path)
    {
        if (ppPathOfFoundObject)
        {
            *ppPathOfFoundObject = pstr_path;
            pstr_path = NULL;
        }
    }
    if (ppTreeNodeEntry)
    {
        *ppTreeNodeEntry = pentry;
        pentry = NULL;
    }

fail:
    SG_STRING_NULLFREE(pCtx, pstr_path);
    SG_TREENODE_ENTRY_NULLFREE(pCtx, pentry);
	SG_NULLFREE(pCtx, psz_hid_treenode);
	SG_VARRAY_NULLFREE(pCtx, pva_gidpaths);
}

void SG_treendx__get_changesets_where_any_of_these_gids_changed(
        SG_context* pCtx, 
        SG_treendx* pTreeNdx, 
        SG_stringarray* psa_gids,
        SG_vhash** ppvh
        )
{
    SG_string* pstr = NULL;
    SG_uint32 count = 0;
    SG_uint32 i = 0;
    SG_vhash* pvh_results = NULL;
    int rc = 0;
    sqlite3_stmt* pStmt = NULL;

    SG_ERR_CHECK(  SG_string__alloc__sz(pCtx, &pstr, "SELECT csid FROM changes WHERE gid IN (")  );
    SG_ERR_CHECK(  SG_stringarray__count(pCtx, psa_gids, &count)  );
    for (i=0; i<count; i++)
    {
        const char* psz_gid = NULL;

        SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, psa_gids, i, &psz_gid)  );

        if (i > 0)
        {
            SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstr, ",")  );
        }
        SG_ERR_CHECK(  SG_string__append__format(pCtx, pstr, "'%s'", psz_gid)  );
    }
    SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstr, ") ORDER BY generation DESC")  );

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_results)  );
	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pTreeNdx->psql, &pStmt, "%s", SG_string__sz(pstr))  );
	while ((rc = sqlite3_step(pStmt)) == SQLITE_ROW)
	{
		const char* psz_csid = (const char*) sqlite3_column_text(pStmt, 0);
        SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh_results, psz_csid)  );

	}
	if (rc != SQLITE_DONE)
	{
		SG_ERR_THROW(SG_ERR_SQLITE(rc));
	}

    *ppvh = pvh_results;
    pvh_results = NULL;

fail:
    SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
    SG_VHASH_NULLFREE(pCtx, pvh_results);
}

void SG_treendx__free(SG_context* pCtx, SG_treendx* pTreeNdx)
{
	if (!pTreeNdx)
		return;

	SG_ERR_IGNORE(  sg_sqlite__close(pCtx, pTreeNdx->psql)  );

	SG_PATHNAME_NULLFREE(pCtx, pTreeNdx->pPath_db);
	SG_NULLFREE(pCtx, pTreeNdx);
}

